from syned.beamline.optical_elements.ideal_elements.lens import IdealLens

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement

from srwlib import SRWLOptL

class SRWIdealLens(IdealLens, SRWOpticalElement):
    def __init__(self, name="Undefined", focal_x=1.0, focal_y=1.0, x=0.0, y=0.0):
        IdealLens.__init__(self, name=name, focal_x=focal_x, focal_y=focal_y)

        self.set_lens_position(x, y)

    def set_lens_position(self, x=0.0, y=0.0):
        self._x = x
        self._y = y

    def toSRWLOpt(self):
        return SRWLOptL(_Fx=self.focal_x(), _Fy=self.focal_y(), _x=self._x, _y=self._y)

    def fromSRWLOpt(self, srwlopt=SRWLOptL()):
        if not isinstance(srwlopt, SRWLOptL):
            raise ValueError("SRW object is not a SRWLOptL object")

        self.__init__(focal_x=srwlopt.Fx, focal_y=srwlopt.Fy, x=srwlopt.x, y=srwlopt.y)

    def to_python_code(self, data=None):
        return data[0] + "="+  "SRWLOptL(_Fx=" + str(self.focal_x()) + ", _Fy=" + str(self.focal_y()) + ", _x=" + str(self._x) + ", _y=" + str(self._y) + ")" + "\n"




from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.magnetic_structures.srw_wiggler import SRWWiggler

class SRWWigglerLightSource(SRWLightSource):

    def __init__(self,
                 name="Undefined",
                 electron_energy_in_GeV = 1.0,
                 electron_energy_spread = 0.0,
                 ring_current = 0.1,
                 electrons_per_bunch = 400,
                 electron_beam_size_h=0.0,
                 electron_beam_size_v=0.0,
                 emittance=0.0,
                 coupling_costant=0.0,
                 K_vertical = 0.0,
                 K_horizontal = 0.0,
                 period_length = 0.0,
                 number_of_periods = 1):

        super().__init__(name,
                         electron_energy_in_GeV = electron_energy_in_GeV,
                         electron_energy_spread = electron_energy_spread,
                         ring_current = ring_current,
                         electrons_per_bunch = electrons_per_bunch,
                         electron_beam_size_h=electron_beam_size_h,
                         electron_beam_size_v=electron_beam_size_v,
                         emittance=emittance,
                         coupling_costant=coupling_costant,
                         magnetic_structure=SRWWiggler(K_vertical,
                                                       K_horizontal,
                                                       period_length,
                                                       number_of_periods))

        self._electron_beam.set_drift_distance(-1.1*self.get_length()/2)

    def get_length(self):
        return self._magnetic_structure._period_length*self._magnetic_structure._number_of_periods




import numpy

from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.light_sources.srw_bending_magnet_light_source import SRWBendingMagnetLightSource
from wofrysrw.storage_ring.srw_light_source import SourceWavefrontParameters, SRWPrecisionParameters

from srxraylib.plot.gol import plot, plot_contour, plot_surface

period_length=0.14

bm = SRWBendingMagnetLightSource("MCX",
                      electron_energy_in_GeV=2.0,
                      electron_energy_spread=0.0007,
                      ring_current=0.4,
                      electron_beam_size_h=0.05545e-3,
                      electron_beam_size_v=2.784e-6,
                      emittance=0.2525e-9,
                      coupling_costant=0.01,
                      magnetic_field=1.2,
                      length=0.2)

print(bm.get_electron_beam().get_electron_beam_geometrical_properties().to_info())

wf_parameters = SourceWavefrontParameters(photon_energy_min = 13000,
                                          photon_energy_max = 13000,
                                          photon_energy_points=1,
                                          h_slit_gap = 20e-3,
                                          v_slit_gap = 2e-3,
                                          h_slit_points=100,
                                          v_slit_points=10,
                                          distance = 10.0,
                                          srw_precision_parameters=SRWPrecisionParameters(relative_precision=0.001))

e, h, v, i = bm.get_intensity(bm.get_SRW_Wavefront(wf_parameters))


plot_contour(i[int(int(e.size/2))],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("MCX",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]",plot_points=0,
             contour_levels=numpy.linspace(0, i.max(), 20), cmap=None, cbar=1,cbar_title="Flux ",show=1)

plot_surface(i[int(e.size/2)],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("MCX",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]",show=1)



import numpy

from srwlib import srwl

from syned.storage_ring.light_source import LightSource

from wofry.beamline.decorators import LightSourceDecorator
from wofrysrw.srw_object import SRWObject
from wofrysrw.storage_ring.srw_magnetic_structure import SRWMagneticStructure
from wofrysrw.storage_ring.srw_electron_beam import SRWElectronBeam, SRWElectronBeamGeometricalProperties
from wofrysrw.propagator.wavefront2D.srw_wavefront import SRWWavefront, WavefrontParameters

class PowerDensityPrecisionParameters(object):
    def __init__(self,
                 precision_factor = 1.5,
                 computation_method = 1, # (1- "near field", 2- "far field")
                 initial_longitudinal_position = 0.0, # (effective if initial_longitudinal_position < final_longitudinal_position)
                 final_longitudinal_position = 0.0, # (effective if initial_longitudinal_position < final_longitudinal_position)
                 number_of_points_for_trajectory_calculation = 20000 #number of points for (intermediate) trajectory calculation
                 ):
        self._precision_factor = precision_factor
        self._computation_method = computation_method
        self._initial_longitudinal_position = initial_longitudinal_position
        self._final_longitudinal_position = final_longitudinal_position
        self._number_of_points_for_trajectory_calculation = number_of_points_for_trajectory_calculation

    def to_SRW_array(self):
        return [float(self._precision_factor),
                int(self._computation_method),
                float(self._initial_longitudinal_position),
                float(self._final_longitudinal_position),
                int(self._number_of_points_for_trajectory_calculation)]

class PhotonSourceProperties(object):

    def __init__(self,
                 rms_h = 0.0,
                 rms_v = 0.0,
                 rms_hp = 0.0,
                 rms_vp = 0.0,
                 coherence_volume_h = 0.0,
                 coherence_volume_v = 0.0,
                 diffraction_limit = 0.0):
        self._rms_h = rms_h
        self._rms_v = rms_v
        self._rms_hp = rms_hp
        self._rms_vp = rms_vp
        self._coherence_volume_h = coherence_volume_h
        self._coherence_volume_v = coherence_volume_v
        self._diffraction_limit = diffraction_limit

    def to_info(self):
        info = 'Photon beam (convolution): \n'
        info += '   RMS size H/V [um]: '+ repr(self._rms_h*1e6) + '  /  ' + repr(self._rms_v*1e6) + '\n'
        info += '   RMS divergence H/V [urad]: '+ repr(self._rms_hp*1e6) + '  /  ' + repr(self._rms_vp*1e6) + '\n\n'
        info += '   Coherent volume in H phase space: '+ repr(self._coherence_volume_h) + '\n'
        info += '   Coherent volume in V phase space: '+ repr(self._coherence_volume_v) + '\n\n'
        info += '   RMS diffraction limit source size [um]: '+ repr(self._diffraction_limit*1e6) + '\n'
        info += '   FWHM diffraction limit source size [um]: '+ repr(self._diffraction_limit*2.35*1e6)

        return info

class SRWLightSource(LightSource, LightSourceDecorator, SRWObject):
    def __init__(self,
                 name="Undefined",
                 electron_beam=SRWElectronBeam(),
                 magnetic_structure=SRWMagneticStructure()):
        LightSource.__init__(self, name, electron_beam, magnetic_structure)

        self.__source_wavefront_parameters = None

    def get_gamma(self):
        return self._electron_beam.gamma()

    def get_photon_source_properties(self):
        return NotImplementedError("must be implemented in subclasses")

    # from Wofry Decorator
    def get_wavefront(self, wavefront_parameters):
        return self.get_SRW_Wavefront(source_wavefront_parameters=wavefront_parameters).toGenericWavefront()

    def get_SRW_Wavefront(self, source_wavefront_parameters = WavefrontParameters()):
        self.__source_wavefront_parameters = source_wavefront_parameters

        mesh = source_wavefront_parameters.to_SRWRadMesh()

        wfr = SRWWavefront()
        wfr.allocate(mesh.ne, mesh.nx, mesh.ny)
        wfr.mesh = mesh
        wfr.partBeam = self._electron_beam.to_SRWLPartBeam()

        srwl.CalcElecFieldSR(wfr,
                             0,
                             self._magnetic_structure.get_SRWLMagFldC(),
                             source_wavefront_parameters._wavefront_precision_parameters.to_SRW_array())

        return wfr

    def get_source_wavefront_parameters(self):
        return self.__source_wavefront_parameters

    def to_python_code(self, data=None):
        is_multi_electron = data

        text_code = self.get_electron_beam().to_python_code()
        text_code += "\n"
        text_code += self.get_magnetic_structure().to_python_code()

        source_wavefront_parameters = self.get_source_wavefront_parameters()

        if not source_wavefront_parameters is None:
            text_code += "\n"
            text_code += source_wavefront_parameters.to_python_code()
            text_code += "\n"
            text_code += "wfr = SRWLWfr()" + "\n"
            text_code += "wfr.allocate(mesh.ne, mesh.nx, mesh.ny)" + "\n"
            text_code += "wfr.mesh = mesh" + "\n"
            text_code += "wfr.partBeam = part_beam" + "\n"
            text_code += "\n"
            text_code += "initial_mesh = deepcopy(wfr.mesh)"
            text_code += "\n"

            if not is_multi_electron:
                text_code += "srwl.CalcElecFieldSR(wfr, 0, magnetic_field_container, "
                text_code += source_wavefront_parameters._wavefront_precision_parameters.to_python_code() + ")" + "\n"

        return text_code

    def get_power_density(self,
                          source_wavefront_parameters = WavefrontParameters(),
                          power_density_precision_parameters = PowerDensityPrecisionParameters()):

        stkP = source_wavefront_parameters.to_SRWLStokes()

        srwl.CalcPowDenSR(stkP,
                          self._electron_beam.to_SRWLPartBeam(),
                          0,
                          self._magnetic_structure.get_SRWLMagFldC(),
                          power_density_precision_parameters.to_SRW_array())

        hArray = numpy.zeros(stkP.mesh.nx)
        vArray = numpy.zeros(stkP.mesh.ny)
        powerArray = numpy.zeros((stkP.mesh.nx,stkP.mesh.ny))

        # fill arrays
        ij = -1
        for j in range(stkP.mesh.ny):
            for i in range(stkP.mesh.nx):
                ij += 1
                xx = stkP.mesh.xStart + i*(stkP.mesh.xFin-stkP.mesh.xStart)/(stkP.mesh.nx-1)
                yy = stkP.mesh.yStart + j*(stkP.mesh.yFin-stkP.mesh.yStart)/(stkP.mesh.ny-1)
                powerArray[i,j] = stkP.arS[ij]
                hArray[i] = xx # mm
                vArray[j] = yy # mm

        return (hArray, vArray, powerArray)

    @classmethod
    def get_total_power_from_power_density(cls, h_array, v_array, power_density_matrix):
        area = (numpy.abs(h_array[1]-h_array[0])*numpy.abs(v_array[1]-v_array[0]))*1e6
        total_power = 0
        for i in range(0, len(h_array)):
            for j in range(0, len(v_array)):
                total_power += power_density_matrix[i, j]*area

        return total_power




from syned.beamline.shape import BoundaryShape

from wofrysrw.beamline.optical_elements.absorbers.srw_slit import SRWSlit

class SRWObstacle(SRWSlit):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        SRWSlit.__init__(self, name=name, boundary_shape=boundary_shape)

    def get_srw_ap_or_ob(self):
       return 'o'


import numpy

from syned.beamline.optical_elements.gratings.grating import Grating
from syned.beamline.shape import Ellipse, Rectangle, Circle

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters

from srwlib import SRWLOptC, SRWLOptMir, SRWLOptG
from srwlib import srwl, srwl_opt_setup_surf_height_1d, srwl_opt_setup_surf_height_2d, srwl_uti_read_data_cols

from wofrysrw.beamline.optical_elements.mirrors.srw_mirror import Orientation, TreatInputOutput, ApertureShape, SimulationMethod

class SRWGrating(Grating, SRWOpticalElement):
    def __init__(self,
                 name                               = "Undefined",
                 tangential_size                    = 1.2,
                 sagittal_size                      = 0.01,
                 grazing_angle                      = 0.003,
                 vertical_position_of_mirror_center = 0.0,
                 horizontal_position_of_mirror_center = 0.0,
                 orientation_of_reflection_plane    = Orientation.UP,
                 invert_tangent_component           = False,
                 height_profile_data_file           = "mirror.dat",
                 height_profile_data_file_dimension = 1,
                 height_amplification_coefficient   = 1.0,
                 diffraction_order                  = 1,
                 grooving_density_0                 =800, # groove density [lines/mm] (coefficient a0 in the polynomial groove density: a0 + a1*y + a2*y^2 + a3*y^3 + a4*y^4)
                 grooving_density_1                 =0.0, # groove density polynomial coefficient a1 [lines/mm^2]
                 grooving_density_2                 =0.0, # groove density polynomial coefficient a2 [lines/mm^3]
                 grooving_density_3                 =0.0, # groove density polynomial coefficient a3 [lines/mm^4]
                 grooving_density_4                 =0.0, # groove density polynomial coefficient a4 [lines/mm^5]
                 grooving_angle                     = 0.0 # angle between the grove direction and the sagittal direction of the substrate
                 ):

        Grating.__init__(self,
                        name=name,
                        boundary_shape=Rectangle(x_left=horizontal_position_of_mirror_center - 0.5*sagittal_size,
                                                 x_right=horizontal_position_of_mirror_center + 0.5*sagittal_size,
                                                 y_bottom=vertical_position_of_mirror_center - 0.5*tangential_size,
                                                 y_top=vertical_position_of_mirror_center + 0.5*tangential_size),
                        surface_shape=self.get_shape(),
                        ruling=grooving_density_0*1e3)

        self.tangential_size                                  = tangential_size
        self.sagittal_size                                    = sagittal_size
        self.grazing_angle                                    = grazing_angle
        self.orientation_of_reflection_plane                  = orientation_of_reflection_plane
        self.invert_tangent_component                         = invert_tangent_component

        self.height_profile_data_file = height_profile_data_file
        self.height_profile_data_file_dimension = height_profile_data_file_dimension
        self.height_amplification_coefficient = height_amplification_coefficient

        self.diffraction_order = diffraction_order

        self.grooving_density_0 = grooving_density_0
        self.grooving_density_1 = grooving_density_1
        self.grooving_density_2 = grooving_density_2
        self.grooving_density_3 = grooving_density_3
        self.grooving_density_4 = grooving_density_4
        self.grooving_angle     = grooving_angle

    def get_alpha_angle(self):
        return self.grazing_angle

    def get_beta_angle(self, photon_energy):
        wavelength = 1.239842e-3/photon_energy # in mm

        return numpy.arcsin(wavelength*self.grooving_density_0 - numpy.cos(self.get_alpha_angle())) # Grating Output Angle

    def get_deflection_angle(self, photon_energy):
        return self.get_alpha_angle() + self.get_beta_angle(photon_energy) + 1.57079632679 # Grating Deflection Angle

    def get_output_orientation_vectors(self, photon_energy):
        deflection_angle = self.get_deflection_angle(photon_energy)
        tangent = 1.0 if not self.invert_tangent_component else -1.0

        if self.orientation_of_reflection_plane == Orientation.UP:
            return 0,  numpy.sin(deflection_angle),  numpy.cos(deflection_angle), tangent, 0.0
        elif self.orientation_of_reflection_plane == Orientation.DOWN:
            return 0,  -numpy.sin(deflection_angle),  numpy.cos(deflection_angle), tangent, 0.0
        elif self.orientation_of_reflection_plane == Orientation.LEFT:
            return numpy.sin(deflection_angle),  0, numpy.cos(deflection_angle), 0.0, tangent
        elif self.orientation_of_reflection_plane == Orientation.RIGHT:
            return -numpy.sin(deflection_angle),  0, numpy.cos(deflection_angle), 0.0, tangent

    def get_shape(self):
        raise NotImplementedError()

    def applyOpticalElement(self, wavefront=None, parameters=None):
        optical_elements = [self.toSRWLOpt()]
        propagation_parameters = [self.get_srw_wavefront_propagation_parameter()]

        if not self.height_profile_data_file is None:
            optical_elements.append(self.get_optTrEr(wavefront))
            propagation_parameters.append(WavefrontPropagationParameters().to_SRW_array())

        optBL = SRWLOptC(optical_elements, propagation_parameters)

        srwl.PropagElecField(wavefront, optBL)

        return wavefront

    def add_to_srw_native_array(self, oe_array = [], pp_array=[], parameters=None, wavefront=None):
        oe_array.append(self.toSRWLOpt())
        pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters))

        if not self.height_profile_data_file is None:
            oe_array.append(self.get_optTrEr(wavefront))
            pp_array.append(WavefrontPropagationParameters().to_SRW_array())

    def get_substrate_mirror(self):
        nvx, nvy, nvz, tvx, tvy = self.get_orientation_vectors()
        x, y = self.getXY()

        if isinstance(self.get_boundary_shape(), Rectangle):
            ap_shape = ApertureShape.RECTANGULAR
        elif isinstance(self.get_boundary_shape(), Ellipse) or isinstance(self.get_boundary_shape(), Circle):
            ap_shape = ApertureShape.ELLIPTIC

        return self.get_SRWLOptMir(nvx, nvy, nvz, tvx, tvy, x, y, ap_shape)


    def toSRWLOpt(self):
        substrate_mirror = self.get_substrate_mirror()

        grating = SRWLOptG(_mirSub=substrate_mirror,
                           _m=self.diffraction_order,
                           _grDen =self.grooving_density_0,
                           _grDen1=self.grooving_density_1,
                           _grDen2=self.grooving_density_2,
                           _grDen3=self.grooving_density_3,
                           _grDen4=self.grooving_density_4,
                           _grAng=self.grooving_angle)

        return grating


    def get_SRWLOptMir(self, nvx, nvy, nvz, tvx, tvy, x, y, ap_shape):
        mirror = SRWLOptMir()

        mirror.set_dim_sim_meth(_size_tang=self.tangential_size,
                                _size_sag=self.sagittal_size,
                                _ap_shape=ap_shape,
                                _sim_meth=SimulationMethod.THICK,
                                _treat_in_out=TreatInputOutput.WAVEFRONT_INPUT_CENTER_OUTPUT_CENTER)
        mirror.set_orient(_nvx=nvx,
                          _nvy=nvy,
                          _nvz=nvz,
                          _tvx=tvx,
                          _tvy=tvy,
                          _x = x,
                          _y = y)

        return mirror

    def get_optTrEr(self, wavefront):
        if self.orientation_of_reflection_plane == Orientation.LEFT or self.orientation_of_reflection_plane == Orientation.RIGHT:
            dim = 'x'
        elif self.orientation_of_reflection_plane == Orientation.UP or self.orientation_of_reflection_plane == Orientation.DOWN:
            dim = 'y'

        if self.height_profile_data_file_dimension == 1:
            height_profile_data = srwl_uti_read_data_cols(self.height_profile_data_file,
                                                          _str_sep='\t',
                                                          _i_col_start=0,
                                                          _i_col_end=1)

            optTrEr = srwl_opt_setup_surf_height_1d(_height_prof_data=height_profile_data,
                                                    _ang=self.grazing_angle,
                                                    _ang_r=self.get_deflection_angle(wavefront.get_photon_energy()),
                                                    _dim=dim,
                                                    _amp_coef=self.height_amplification_coefficient)

        elif self.height_profile_data_file_dimension == 2:
            height_profile_data = srwl_uti_read_data_cols(self.height_profile_data_file,
                                                          _str_sep='\t')

            optTrEr = srwl_opt_setup_surf_height_2d(_height_prof_data=height_profile_data,
                                                    _ang=self.grazing_angle,
                                                    _ang_r=self.get_deflection_angle(wavefront.get_photon_energy()),
                                                    _dim=dim,
                                                    _amp_coef=self.height_amplification_coefficient)

        return optTrEr



    def to_python_code(self, data=None):
        oe_name = data[0]
        wavefront = data[1]

        nvx, nvy, nvz, tvx, tvy = self.get_orientation_vectors()
        x, y = self.getXY()

        if isinstance(self.get_boundary_shape(), Rectangle):
            ap_shape = ApertureShape.RECTANGULAR
        elif isinstance(self.get_boundary_shape(), Ellipse) or isinstance(self.get_boundary_shape(), Circle):
            ap_shape = ApertureShape.ELLIPTIC

        text_code = self.to_python_code_aux(nvx, nvy, nvz, tvx, tvy, x, y, ap_shape)

        text_code += "substrate_mirror.set_dim_sim_meth(_size_tang=" + str(self.tangential_size) + "," + "\n"
        text_code += "                                  _size_sag=" + str(self.sagittal_size) + "," + "\n"
        text_code += "                                  _ap_shape='" + str(ap_shape) + "'," + "\n"
        text_code += "                                  _sim_meth=" + str(SimulationMethod.THICK) + "," + "\n"
        text_code += "                                  _treat_in_out=" + str(TreatInputOutput.WAVEFRONT_INPUT_CENTER_OUTPUT_CENTER) + ")" + "\n"

        text_code += "substrate_mirror.set_orient(_nvx=" + str(nvx) + "," + "\n"
        text_code += "                            _nvy=" + str(nvy) + "," + "\n"
        text_code += "                            _nvz=" + str(nvz) + "," + "\n"
        text_code += "                            _tvx=" + str(tvx) + "," + "\n"
        text_code += "                            _tvy=" + str(tvy) + "," + "\n"
        text_code += "                            _x=" + str(x) + "," + "\n"
        text_code += "                            _y=" + str(y) + ")" + "\n"

        text_code += "\n"

        text_code += oe_name + "="+ "SRWLOptG(_mirSub=substrate_mirror" + "," + "\n"
        text_code += "            _m="     + str(self.diffraction_order) + "," + "\n"
        text_code += "            _grDen=" + str(self.grooving_density_0) + "," + "\n"
        text_code += "            _grDen1="+ str(self.grooving_density_1) + "," + "\n"
        text_code += "            _grDen2="+ str(self.grooving_density_2) + "," + "\n"
        text_code += "            _grDen3="+ str(self.grooving_density_3) + "," + "\n"
        text_code += "            _grDen4="+ str(self.grooving_density_4) + "," + "\n"
        text_code += "            _grAng= "+ str(self.grooving_angle) + ")" + "\n"

        if not self.height_profile_data_file is None:
            text_code += "\n"

            if self.orientation_of_reflection_plane == Orientation.LEFT or self.orientation_of_reflection_plane == Orientation.RIGHT:
                dim = 'x'
            elif self.orientation_of_reflection_plane == Orientation.UP or self.orientation_of_reflection_plane == Orientation.DOWN:
                dim = 'y'

            if self.height_profile_data_file_dimension == 1:
                text_code += "height_profile_data = srwl_uti_read_data_cols('" + self.height_profile_data_file + "'," + "\n"
                text_code += "                                              _str_sep='\\t'," + "\n"
                text_code += "                                              _i_col_start=0," + "\n"
                text_code += "                                              _i_col_end=1)" + "\n"

                text_code += "optTrEr_" + oe_name + " = srwl_opt_setup_surf_height_1d(_height_prof_data=height_profile_data," + "\n"
                text_code += "                                                        _ang="+ str(self.grazing_angle) + "," + "\n"
                text_code += "                                                        _ang_r="+ str(self.get_deflection_angle(wavefront.get_photon_energy())) + "," + "\n"
                text_code += "                                                        _dim='"+ dim + "'," + "\n"
                text_code += "                                                        _amp_coef="+ str(self.height_amplification_coefficient) + ")" + "\n"

            elif self.height_profile_data_file_dimension == 2:
                text_code += "height_profile_data = srwl_uti_read_data_cols('" + self.height_profile_data_file + "'," + "\n"
                text_code += "                                              _str_sep=\'\\t\')" + "\n"

                text_code += "optTrEr_" + oe_name + " = srwl_opt_setup_surf_height_2d(_height_prof_data=height_profile_data," + "\n"
                text_code += "                                                        _ang="+ str(self.grazing_angle) + "," + "\n"
                text_code += "                                                        _ang_r="+ str(self.get_deflection_angle(wavefront.get_photon_energy())) + "," + "\n"
                text_code += "                                                        _dim='"+ dim + "'," + "\n"
                text_code += "                                                        _amp_coef="+ str(self.height_amplification_coefficient) + ")" + "\n"

        return text_code



from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement
from wofrysrw.beamline.optical_elements.mirrors.srw_mirror import SRWMirror, Orientation, ApertureShape, SimulationMethod, TreatInputOutput
from syned.beamline.shape import Sphere

from srwlib import SRWLOptMirSph

class SRWSphericalMirror(SRWMirror, SRWOpticalElement):
    def __init__(self,
                 name                                       = "Undefined",
                 tangential_size                            = 1.2,
                 sagittal_size                              = 0.01,
                 grazing_angle                              = 0.003,
                 orientation_of_reflection_plane            = Orientation.UP,
                 invert_tangent_component                   = False,
                 radius                                     = 1,
                 height_profile_data_file                   = "mirror.dat",
                 height_profile_data_file_dimension         = 1,
                 height_amplification_coefficient           = 1.0):

        super().__init__(name=name,
                         tangential_size=tangential_size,
                         sagittal_size=sagittal_size,
                         grazing_angle=grazing_angle,
                         orientation_of_reflection_plane=orientation_of_reflection_plane,
                         invert_tangent_component=invert_tangent_component,
                         height_profile_data_file=height_profile_data_file,
                         height_profile_data_file_dimension=height_profile_data_file_dimension,
                         height_amplification_coefficient=height_amplification_coefficient)

        self.radius  = radius

    def get_shape(self):
        return Sphere()

    def get_SRWLOptMir(self, nvx, nvy, nvz, tvx, tvy, x, y, ap_shape):
        return SRWLOptMirSph(_size_tang=self.tangential_size,
                            _size_sag=self.sagittal_size,
                            _r=self.radius,
                            _ap_shape=ap_shape,
                            _sim_meth=SimulationMethod.THICK,
                            _treat_in_out=TreatInputOutput.WAVEFRONT_INPUT_CENTER_OUTPUT_CENTER,
                            _nvx=nvx,
                            _nvy=nvy,
                            _nvz=nvz,
                            _tvx=tvx,
                            _tvy=tvy,
                            _x=x,
                            _y=y)

    def fromSRWLOpt(self, srwlopt=SRWLOptMirSph()):
        if not isinstance(srwlopt, SRWLOptMirSph):
            raise ValueError("SRW object is not a SRWLOptMirEl object")

        super().fromSRWLOpt(srwlopt)

        self.radius = srwlopt.rad

    def to_python_code_aux(self, nvx, nvy, nvz, tvx, tvy, x, y, ap_shape):
        text_code  = "SRWLOptMirSph(_size_tang=" + str(self.tangential_size) +"," + "\n"
        text_code += "                   _size_sag=" + str(self.sagittal_size) +"," + "\n"
        text_code += "                   _r=" + str(self.radius) +"," + "\n"
        text_code += "                   _ap_shape='" + str(ap_shape) +"'," + "\n"
        text_code += "                   _sim_meth=" + str(SimulationMethod.THICK) +"," + "\n"
        text_code += "                   _treat_in_out=" + str(TreatInputOutput.WAVEFRONT_INPUT_CENTER_OUTPUT_CENTER) +"," + "\n"
        text_code += "                   _nvx=" + str(nvx) +"," + "\n"
        text_code += "                   _nvy=" + str(nvy) +"," + "\n"
        text_code += "                   _nvz=" + str(nvz) +"," + "\n"
        text_code += "                   _tvx=" + str(tvx) +"," + "\n"
        text_code += "                   _tvy=" + str(tvy) +"," + "\n"
        text_code += "                   _x=" + str(x) +"," + "\n"
        text_code += "                   _y=" + str(y) +")" + "\n"

        return text_code



from syned.beamline.optical_elements.ideal_elements.screen import Screen

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement

class SRWScreen(Screen, SRWOpticalElement):
    def __init__(self, name="Undefined"):
        Screen.__init__(self, name=name)

    def applyOpticalElement(self, wavefront=None, parameters=None):
        return wavefront

    def add_to_srw_native_array(self, oe_array = [], pp_array=[], parameters=None, wavefront=None):
        pass

    def to_python_code(self, data=None):
        return ""



import numpy

from syned.beamline.shape import Ellipse, Rectangle, Circle
from wofry.beamline.decorators import OpticalElementDecorator

from wofry.propagator.propagator import PropagationParameters
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters, WavefrontPropagationOptionalParameters
from wofrysrw.srw_object import SRWObject

from srwlib import SRWLOptC, srwl

class Orientation:
    UP = 0
    DOWN = 1
    LEFT = 2
    RIGHT = 3

class SRWOpticalElementDecorator(SRWObject):
    def toSRWLOpt(self):
        raise NotImplementedError("")

    @classmethod
    def fromSRWLOpt(cls, srwlopt=None):
        raise NotImplementedError("")

class SRWOpticalElement(SRWOpticalElementDecorator, OpticalElementDecorator):

    def applyOpticalElement(self, wavefront=None, parameters=None):
        optBL = SRWLOptC([self.toSRWLOpt()],
                         [self.get_srw_wavefront_propagation_parameter(parameters)])

        srwl.PropagElecField(wavefront, optBL)

        return wavefront

    def add_to_srw_native_array(self, oe_array = [], pp_array=[], parameters=None, wavefront=None):
        oe_array.append(self.toSRWLOpt())
        pp_array.append(self.get_srw_wavefront_propagation_parameter(parameters))

    def get_srw_wavefront_propagation_parameter(self, parameters):
        if isinstance(parameters, PropagationParameters):
            if not parameters.has_additional_parameter("srw_oe_wavefront_propagation_parameters"):
                wavefront_propagation_parameters = WavefrontPropagationParameters()
            else:
                wavefront_propagation_parameters = parameters.get_additional_parameter("srw_oe_wavefront_propagation_parameters")

                if not isinstance(wavefront_propagation_parameters, WavefrontPropagationParameters):
                    raise ValueError("SRW Wavefront Propagation Parameters are inconsistent")

            srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()

            if parameters.has_additional_parameter("srw_oe_wavefront_propagation_optional_parameters"):
                wavefront_propagation_optional_parameters = parameters.get_additional_parameter("srw_oe_wavefront_propagation_optional_parameters")

                if not isinstance(wavefront_propagation_optional_parameters, WavefrontPropagationOptionalParameters):
                    raise ValueError("SRW Wavefront Propagation Optional Parameters are inconsistent")

                wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)
        elif isinstance(parameters, list):
            wavefront_propagation_parameters = parameters[0]
            wavefront_propagation_optional_parameters = parameters[1]

            srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()
            if not wavefront_propagation_optional_parameters is None: wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)

        return srw_parameters_array

    def getXY(self):
        boundary_shape = self.get_boundary_shape()

        if not boundary_shape is None:
            if isinstance(boundary_shape, Rectangle) or isinstance(boundary_shape, Ellipse):
                x_left, x_right, y_bottom, y_top = boundary_shape.get_boundaries()

                return x_left + 0.5*(x_right-x_left), \
                       y_bottom + 0.5*(y_top-y_bottom)

            elif isinstance(boundary_shape, Circle):
                radius, x_center, y_center = boundary_shape.get_boundaries()

                return x_center, y_center
        else:
            return 0.0, 0.0

    def get_orientation_vectors(self):
        sign = (-1 if self.invert_tangent_component else 1)

        if self.orientation_of_reflection_plane == Orientation.LEFT:
            nvx = -numpy.cos(self.grazing_angle)
            nvy = 0
            nvz = -numpy.sin(self.grazing_angle)
            tvx = sign*numpy.sin(self.grazing_angle)
            tvy = 0
        elif self.orientation_of_reflection_plane == Orientation.RIGHT:
            nvx = numpy.cos(self.grazing_angle)
            nvy = 0
            nvz = -numpy.sin(self.grazing_angle)
            tvx = -sign*numpy.sin(self.grazing_angle)
            tvy = 0
        elif self.orientation_of_reflection_plane == Orientation.UP:
            nvx = 0
            nvy = numpy.cos(self.grazing_angle)
            nvz = -numpy.sin(self.grazing_angle)
            tvx = 0
            tvy = -sign*numpy.sin(self.grazing_angle)
        elif self.orientation_of_reflection_plane == Orientation.DOWN:
            nvx = 0
            nvy = -numpy.cos(self.grazing_angle)
            nvz = -numpy.sin(self.grazing_angle)
            tvx = 0
            tvy = -sign*numpy.sin(self.grazing_angle)

        return nvx, nvy, nvz, tvx, tvy



from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.shape import BoundaryShape, Rectangle, Ellipse

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement

from srwlib import SRWLOptA

class SRWSlit(Slit, SRWOpticalElement):
    def __init__(self, name="Undefined", boundary_shape=BoundaryShape()):
        Slit.__init__(self, name=name, boundary_shape=boundary_shape)

    def toSRWLOpt(self):
        boundaries = self._boundary_shape.get_boundaries()

        Dx = abs(boundaries[1]-boundaries[0])
        Dy = abs(boundaries[3]-boundaries[2])
        x = 0.5*(boundaries[1]+boundaries[0])
        y = 0.5*(boundaries[3]+boundaries[2])

        if isinstance(self._boundary_shape, Rectangle):
            shape = 'r'
        elif isinstance(self._boundary_shape, Ellipse):
            if Dx != Dy:
                raise ValueError("SRW doesn't support elliptic apertures")

            shape = 'c'

        return SRWLOptA(_shape=shape,
                        _ap_or_ob=self.get_srw_ap_or_ob(),
                        _Dx=Dx,
                        _Dy=Dy,
                        _x=x,
                        _y=y)


    def fromSRWLOpt(self, srwlopt=SRWLOptA()):
        if not isinstance(srwlopt, SRWLOptA):
            raise ValueError("SRW object is not a SRWLOptA object")
        
        if not srwlopt.ap_or_ob == self.get_srw_ap_or_ob():
            raise ValueError("SRW object not compatible")

        if srwlopt.shape == 'r':
            boundary_shape = Rectangle(x_left=srwlopt.x - 0.5 * srwlopt.Dx,
                                       x_right=srwlopt.x + 0.5 * srwlopt.Dx,
                                       y_bottom=srwlopt.y - 0.5 * srwlopt.Dy,
                                       y_top=srwlopt.y + 0.5 * srwlopt.Dy)
        elif srwlopt.shape == 'c':
            boundary_shape = Ellipse(min_ax_left=srwlopt.x - 0.5 * srwlopt.Dx,
                                     min_ax_right=srwlopt.x + 0.5 * srwlopt.Dx,
                                     maj_ax_bottom=srwlopt.y - 0.5 * srwlopt.Dy,
                                     maj_ax_top=srwlopt.y + 0.5 * srwlopt.Dy)

        self.__init__(boundary_shape=boundary_shape)

    def get_srw_ap_or_ob(self):
        raise NotImplementedError()

    def to_python_code(self, data=None):
        oe_name = data[0]

        boundaries = self._boundary_shape.get_boundaries()

        Dx = abs(boundaries[1]-boundaries[0])
        Dy = abs(boundaries[3]-boundaries[2])
        x = 0.5*(boundaries[1]+boundaries[0])
        y = 0.5*(boundaries[3]+boundaries[2])

        if isinstance(self._boundary_shape, Rectangle):
            shape = 'r'
        elif isinstance(self._boundary_shape, Ellipse):
            if Dx != Dy:
                raise ValueError("SRW doesn't support elliptic apertures")

            shape = 'c'

        text_code  = oe_name + "="+  "SRWLOptA(_shape='" + shape + "'," + "\n"
        text_code += "            _ap_or_ob='" + self.get_srw_ap_or_ob() + "'," + "\n"
        text_code += "            _Dx=" + str(Dx) + "," + "\n"
        text_code += "            _Dy=" + str(Dy) + "," + "\n"
        text_code += "            _x="  + str(x) + "," + "\n"
        text_code += "            _y="  + str(y) + ")" + "\n"

        return text_code



from syned.storage_ring.magnetic_structure import MagneticStructure
from srwlib import array, SRWLMagFldC

from wofrysrw.srw_object import SRWObject

class SRWMagneticStructureDecorator():

    def get_SRWMagneticStructure(self):
        raise NotImplementedError("this method should be implented in subclasses")

    def get_SRWLMagFldC(self):
        return SRWLMagFldC([self.get_SRWMagneticStructure()], array('d', [0]), array('d', [0]), array('d', [0]))


class SRWMagneticStructure(SRWMagneticStructureDecorator, SRWObject):
    def __init__(self):
        super().__init__()

    def to_python_code(self, data=None):
        text_code  = self.to_python_code_aux()
        text_code += "magnetic_field_container = SRWLMagFldC([magnetic_structure], array('d', [0]), array('d', [0]), array('d', [0]))"  + "\n"

        return text_code

    def to_python_code_aux(self):
        raise NotImplementedError()



from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontParameters, SRWWavefront
from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.srw_electron_beam import SRWElectronBeam

from srwlib import *

'''
x = 0.0, #Transverse Coordinates of Gaussian Beam Center at Waist [m]
y = 0.0,
z = 0.0, #Longitudinal Coordinate of Waist [m]
xp = 0.0,  #Average Angles of Gaussian Beam at Waist [rad]
yp = 0.0,
avgPhotEn = 12400, #5000 #Photon Energy [eV]
pulseEn = 0.001, #Energy per Pulse [J] - to be corrected
repRate = 1, #Rep. Rate [Hz] - to be corrected
polar = 1, #1- linear hoirizontal
sigX = 23e-06/2.35, #Horiz. RMS size at Waist [m]
sigY = 23e-06/2.35, #Vert. RMS size at Waist [m]
sigT = 10e-15, #Pulse duration [s] (not used?)
mx = 0, #Transverse Gauss-Hermite Mode Orders
my = 0
'''

class Polarization:
    LINEAR_HORIZONTAL  = 1
    LINEAR_VERTICAL    = 2
    LINEAR_45_DEGREES  = 3
    LINEAR_135_DEGREES = 4
    CIRCULAR_RIGHT     = 5
    CIRCULAR_LEFT      = 6

    @classmethod
    def tuple(cls):
        return ["Linear Horizontal",
                "Linear Vertical",
                "Linear 45\u00b0",
                "Linear 135\u00b0",
                "Circular Right",
                "Circular Left"]

class SRWGaussianLightSource(SRWLightSource):

    def __init__(self,
                 name="Undefined",
                 beam_center_at_waist_x = 0.0, #Transverse Coordinates of Gaussian Beam Center at Waist [m]
                 beam_center_at_waist_y = 0.0,
                 beam_center_at_waist_z = 0.0, #Longitudinal Coordinate of Waist [m]
                 average_angle_at_waist_x = 0.0, #Average Angles of Gaussian Beam at Waist [rad]
                 average_angle_at_waist_y = 0.0,
                 photon_energy = 12400,
                 energy_per_pulse = 0.001, #Energy per Pulse [J]
                 repetition_rate = 1, #[Hz]
                 polarization = Polarization.LINEAR_HORIZONTAL,
                 horizontal_sigma_at_waist = 1e-6,
                 vertical_sigma_at_waist = 1e-6,
                 pulse_duration = 10e-15, #[s]
                 transverse_gauss_hermite_mode_order_x = 0,
                 transverse_gauss_hermite_mode_order_y = 0
                 ):
        super().__init__(name,
                         electron_beam=None,
                         magnetic_structure=None)

        self.beam_center_at_waist_x                = beam_center_at_waist_x
        self.beam_center_at_waist_y                = beam_center_at_waist_y
        self.beam_center_at_waist_z                = beam_center_at_waist_z
        self.average_angle_at_waist_x              = average_angle_at_waist_x
        self.average_angle_at_waist_y              = average_angle_at_waist_y
        self.photon_energy                         = photon_energy
        self.energy_per_pulse                      = energy_per_pulse
        self.repetition_rate                       = repetition_rate
        self.polarization                          = polarization
        self.horizontal_sigma_at_waist             = horizontal_sigma_at_waist
        self.vertical_sigma_at_waist               = vertical_sigma_at_waist
        self.pulse_duration                        = pulse_duration
        self.transverse_gauss_hermite_mode_order_x = transverse_gauss_hermite_mode_order_x
        self.transverse_gauss_hermite_mode_order_y = transverse_gauss_hermite_mode_order_y

    # from Wofry Decorator
    def get_wavefront(self, wavefront_parameters):
        return self.get_SRW_Wavefront(source_wavefront_parameters=wavefront_parameters).toGenericWavefront()

    def get_SRW_Wavefront(self, source_wavefront_parameters = WavefrontParameters()):
        self.__source_wavefront_parameters = source_wavefront_parameters

        source_wavefront_parameters.photon_energy_min = self.photon_energy
        source_wavefront_parameters.photon_energy_max = self.photon_energy
        source_wavefront_parameters.photon_energy_points = 1

        mesh = source_wavefront_parameters.to_SRWRadMesh()

        GsnBm = SRWLGsnBm() #Gaussian Beam structure (just parameters)
        GsnBm.x         = self.beam_center_at_waist_x
        GsnBm.y         = self.beam_center_at_waist_y
        GsnBm.z         = self.beam_center_at_waist_z
        GsnBm.xp        = self.average_angle_at_waist_x
        GsnBm.yp        = self.average_angle_at_waist_y
        GsnBm.avgPhotEn = self.photon_energy
        GsnBm.pulseEn   = self.energy_per_pulse
        GsnBm.repRate   = self.repetition_rate
        GsnBm.polar     = self.polarization
        GsnBm.sigX      = self.horizontal_sigma_at_waist
        GsnBm.sigY      = self.vertical_sigma_at_waist
        GsnBm.sigT      = self.pulse_duration
        GsnBm.mx        = self.transverse_gauss_hermite_mode_order_x
        GsnBm.my        = self.transverse_gauss_hermite_mode_order_y

        wfr = SRWWavefront()
        wfr.allocate(mesh.ne, mesh.nx, mesh.ny)
        wfr.mesh = mesh

        wfr.partBeam.partStatMom1.x = GsnBm.x
        wfr.partBeam.partStatMom1.y = GsnBm.y
        wfr.partBeam.partStatMom1.z = GsnBm.z
        wfr.partBeam.partStatMom1.xp = GsnBm.xp
        wfr.partBeam.partStatMom1.yp = GsnBm.yp

        arPrecPar = [source_wavefront_parameters._wavefront_precision_parameters._sampling_factor_for_adjusting_nx_ny]

        srwl.CalcElecFieldGaussian(wfr, GsnBm, arPrecPar)

        return wfr

    def get_source_wavefront_parameters(self):
        return self.__source_wavefront_parameters

    def to_python_code(self, data=None):
        is_multi_electron = data

        text_code = ""

        source_wavefront_parameters = self.get_source_wavefront_parameters()

        if not source_wavefront_parameters is None:
            text_code += source_wavefront_parameters.to_python_code()
            text_code += "\n"
            text_code += "wfr = SRWLWfr()" + "\n"
            text_code += "wfr.allocate(mesh.ne, mesh.nx, mesh.ny)" + "\n"
            text_code += "wfr.mesh = mesh" + "\n"
            text_code += "\n"
            text_code += "initial_mesh = deepcopy(wfr.mesh)" + "\n"
            text_code += "\n"
            text_code += "GsnBm = SRWLGsnBm()" + "\n"
            text_code += "GsnBm.x         = " + str(self.beam_center_at_waist_x) + "\n"
            text_code += "GsnBm.y         = " + str(self.beam_center_at_waist_y) + "\n"
            text_code += "GsnBm.z         = " + str(self.beam_center_at_waist_z) + "\n"
            text_code += "GsnBm.xp        = " + str(self.average_angle_at_waist_x) + "\n"
            text_code += "GsnBm.yp        = " + str(self.average_angle_at_waist_y) + "\n"
            text_code += "GsnBm.avgPhotEn = " + str(self.photon_energy) + "\n"
            text_code += "GsnBm.pulseEn   = " + str(self.energy_per_pulse) + "\n"
            text_code += "GsnBm.repRate   = " + str(self.repetition_rate) + "\n"
            text_code += "GsnBm.polar     = " + str(self.polarization) + "\n"
            text_code += "GsnBm.sigX      = " + str(self.horizontal_sigma_at_waist) + "\n"
            text_code += "GsnBm.sigY      = " + str(self.vertical_sigma_at_waist) + "\n"
            text_code += "GsnBm.sigT      = " + str(self.pulse_duration) + "\n"
            text_code += "GsnBm.mx        = " + str(self.transverse_gauss_hermite_mode_order_x) + "\n"
            text_code += "GsnBm.my        = " + str(self.transverse_gauss_hermite_mode_order_y) + "\n"
            text_code += "\n"
            text_code += "wfr.partBeam.partStatMom1.x = GsnBm.x" + "\n"
            text_code += "wfr.partBeam.partStatMom1.y = GsnBm.y" + "\n"
            text_code += "wfr.partBeam.partStatMom1.z = GsnBm.z" + "\n"
            text_code += "wfr.partBeam.partStatMom1.xp = GsnBm.xp" + "\n"
            text_code += "wfr.partBeam.partStatMom1.yp = GsnBm.yp" + "\n"
            text_code += "\n"

            if not is_multi_electron:
                text_code += "srwl.CalcElecFieldGaussian(wfr, GsnBm, [" + str(source_wavefront_parameters._wavefront_precision_parameters._sampling_factor_for_adjusting_nx_ny) + "])" + "\n"

        return text_code




# THIS FILE IS GENERATED FROM wofrysrw SETUP.PY
short_version = '1.0.16'
version = '1.0.16'
full_version = '1.0.16.dev0+f46b337'
git_revision = 'f46b33709ee9592284919582f610d37af7448970'
release = False

if not release:
    version = full_version
    short_version += ".dev"



from srwlib import SRWLMagFldH, SRWLMagFldU

from syned.storage_ring.magnetic_structures.undulator import Undulator
from wofrysrw.storage_ring.srw_magnetic_structure import SRWMagneticStructure

class SRWUndulator(Undulator, SRWMagneticStructure):

    def __init__(self,
                 K_vertical = 0.0,
                 K_horizontal = 0.0,
                 period_length = 0.0,
                 number_of_periods = 1):
        Undulator.__init__(self, K_vertical, K_horizontal, period_length, number_of_periods)

    def get_SRWMagneticStructure(self):
        magnetic_fields = []

        if self._K_vertical > 0.0:
            magnetic_fields.append(SRWLMagFldH(1, 'v', self.magnetic_field_vertical(), 0, 1, 1))

        if self._K_horizontal > 0.0:
            magnetic_fields.append(SRWLMagFldH(1, 'h', self.magnetic_field_horizontal(), 0, -1, 1))

        return SRWLMagFldU(magnetic_fields,
                           self._period_length,
                           self._number_of_periods)

    def to_python_code_aux(self):
        text_code = "magnetic_fields = []" + "\n"

        if self._K_vertical > 0.0:
            text_code += "magnetic_fields.append(SRWLMagFldH(1, 'v', " + str(self.magnetic_field_vertical()) + ", 0, 1, 1))" + "\n"

        if self._K_horizontal > 0.0:
            text_code += "magnetic_fields.append(SRWLMagFldH(1, 'h', " + str(self.magnetic_field_horizontal()) + ", 0, -1, 1))" + "\n"

        text_code += "magnetic_structure = SRWLMagFldU(magnetic_fields," + str(self._period_length) + "," + str(self._number_of_periods) + ")" + "\n"

        return text_code
